import logging

from app import settings

import eventy.config
import eventy.config.sanic
from eventy.logging import GkeHandler, SimpleHandler

# Configure Eventy
eventy.config.SERVICE_NAME = 'test-sanic-integration'
eventy.config.sanic.SANIC_ACCESS_HEALTH_ROUTE = '/health'
eventy.config.sanic.SANIC_ACCESS_DISABLE_HEALTH_LOGGING = settings.SKIP_HEALTHCHECK_LOGGING

# Setup logging
root_logger = logging.getLogger()
root_logger.setLevel(settings.LOG_LEVEL)

root_handler: logging.Handler

if settings.GKE_LOGGING:
    root_handler = GkeHandler()
else:
    root_handler = SimpleHandler(colored=settings.COLORED_LOGGING)

root_logger.addHandler(root_handler)



# Copyright (c) Qotto, 2021

"""
Eventy sanic integration configuration
"""

__all__ = [
    'SANIC_ACCESS_DISABLE_HEALTH_LOGGING',
    'SANIC_ACCESS_HEALTH_ROUTE',
]

SANIC_ACCESS_DISABLE_HEALTH_LOGGING: bool = True
"""
Disable sanic access logs for health checks
"""

SANIC_ACCESS_HEALTH_ROUTE: str = '/health'
"""
Route for health checks
"""



# Copyright (c) Qotto, 2021

"""
Provides logging handlers to integrate tracing in log messages.

Include a :class:`simple_handler.SimpleHandler`
and a Google Kubernetes Engines specific :class:`gke_handler.GkeHandler`
"""

from .gke_handler import GkeHandler
from .info_providers import correlation_id_provider, request_id_provider
from .simple_handler import SimpleHandler

__all__ = [
    'correlation_id_provider',
    'request_id_provider',
    'SimpleHandler',
    'GkeHandler',
]



# Copyright (c) Qotto, 2021

"""
Django integration utilities

Utility functions to integrate the eventy protocol in django apps
"""
import logging
from typing import Callable

import django.http as vendor_django_http

import eventy.config.django
from eventy.trace_id import correlation_id_var, request_id_var
from eventy.trace_id.generator import gen_trace_id

logger = logging.getLogger(__name__)


def django_trace_middleware(
    get_response: Callable[[vendor_django_http.HttpRequest], vendor_django_http.HttpResponse]
) -> Callable:
    """
    Middleware to extract and propagate correlation_id and generate request_id.

    Log each access, except for health check on ``eventy.config.django.DJANGO_ACCESS_HEALTH_ROUTE``
    if ``eventy.config.django.DJANGO_ACCESS_DISABLE_HEALTH_LOGGING`` is set to True

    :param get_response: Actual response logic
    :return: Modified response
    """

    def _middleware(request: vendor_django_http.HttpRequest) -> vendor_django_http.HttpResponse:
        # test if access should be logged
        is_health_check: bool = request.method == 'GET' and request.path == eventy.config.django.DJANGO_ACCESS_HEALTH_ROUTE
        log_access: bool = not (is_health_check and eventy.config.django.DJANGO_ACCESS_DISABLE_HEALTH_LOGGING)

        # fetch or generate correlation_id
        correlation_id = None
        for key, val in request.headers.items():
            if key.lower() == 'x-correlation-id':
                correlation_id = val
        correlation_id_var.set(correlation_id or gen_trace_id(f'{request.method}_{request.path}'))

        # always generate request_id
        request_id_var.set(gen_trace_id(f'{request.method}_{request.path}'))

        # log request
        if log_access:
            logger.info(f'Request: {request.method} {request.path}')

        # get response
        response = get_response(request)

        # log response
        if log_access:
            logger.info(f'Response: {request.method} {request.path} {response.status_code} ({len(response.content)})')

        return response

    logger.info("Eventy for Django middleware initialized.")
    return _middleware



# Copyright (c) Qotto, 2021

"""
Transactional messaging API (Transactional Producer and Processor)
"""

from .base import Consumer
from eventy.record import Record

__all__ = [
    'TransactionalProducer',
    'TransactionalProcessor',
]


class TransactionalProducer:
    """
    Interface for a transactional record producer
    """

    def add_to_transaction(self, destination: str, record: Record, *arg, **kwargs) -> None:
        """
        Add a record to be committed later

        :param destination: Destination of the record
        :param record: Record to be sent
        """
        raise NotImplementedError

    def commit(self, *arg, **kwargs):
        """
        Produce messages added to transaction in an atomic operation
        """
        raise NotImplementedError


class TransactionalProcessor(Consumer, TransactionalProducer):
    """
    Produce records and commit read atomically
    """

    def commit(self, *arg, **kwargs):
        """
        Produce messages added to transaction and commit read in an atomic operation
        """
        raise NotImplementedError



import logging
import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..', '..', '..'))
sys.path.append(os.path.join(os.path.dirname(__file__)))

import eventy.config
import eventy.config.django
from eventy.logging import GkeHandler, SimpleHandler
from app import settings

# Configure Eventy
eventy.config.SERVICE_NAME = 'eventy-django-test'
# Django integration
eventy.config.django.DJANGO_ACCESS_HEALTH_ROUTE = '/health'
eventy.config.django.DJANGO_ACCESS_DISABLE_HEALTH_LOGGING = settings.SKIP_HEALTHCHECK_LOGGING

# Setup logging
root_logger = logging.getLogger()
root_logger.setLevel(settings.LOG_LEVEL)

root_handler: logging.Handler

if settings.GKE_LOGGING:
    root_handler = GkeHandler()
else:
    root_handler = SimpleHandler(colored=settings.COLORED_LOGGING)

root_logger.addHandler(root_handler)
# logging.getLogger('django.server').addHandler(root_handler)




# Copyright (c) Qotto, 2021

"""
Sanic integration utilities

Utility functions to integrate the eventy protocol in sanic apps.

You need to install sanic optional dependencies ("pip install eventy[sanic]").
"""
import logging
from logging import LogRecord
from typing import Optional
from urllib.parse import urlparse

import sanic.app as vendor_sanic_app
import sanic.request as vendor_sanic_request

import eventy.config.sanic
from eventy.trace_id import correlation_id_var, request_id_var
from eventy.trace_id.generator import gen_trace_id

__all__ = [
    'sanic_correlation_id_provider',
    'sanic_request_id_provider',
    'sanic_access_log_ignore_health_filter',
    'sanic_access_log_msg_provider',
    'Sanic',
]

logger = logging.getLogger(__name__)


def sanic_access_log_ignore_health_filter(record: LogRecord) -> bool:
    """
    Ignore logging for health checks on route /health.

    :param record: Log record to filter
    :return: True only if the log record is a sanic.access log for the /health route
    """
    if not eventy.config.sanic.SANIC_ACCESS_DISABLE_HEALTH_LOGGING:
        return True
    req: str = getattr(record, 'request', None)
    status: int = getattr(record, 'status', None)
    if status != 200:
        return True
    if not req.startswith('GET '):
        return True
    if not req.endswith(eventy.config.sanic.SANIC_ACCESS_HEALTH_ROUTE):
        return True
    if not urlparse(req[4:]).path == eventy.config.sanic.SANIC_ACCESS_HEALTH_ROUTE:
        return True
    return False


def sanic_access_log_msg_provider(record: LogRecord) -> bool:
    """
    Create the ``msg`` field in the log record.

    :param record: Log record to annotate
    :return: True (not really a filter)
    """
    record.msg = f'{record.request} {record.status} ({record.byte})'  # type: ignore
    return True


def sanic_correlation_id_provider(request: vendor_sanic_request.Request) -> None:
    """
    Fetch the correlation_id from the request headers or generate a new one.

    :param request: Sanic request
    """
    correlation_id: Optional[str] = None
    for header_name, header_value in request.headers.items():
        if header_name.lower() == 'x-correlation-id':
            correlation_id = header_value
    correlation_id_var.set(correlation_id or gen_trace_id(request.path))


def sanic_request_id_provider(request: vendor_sanic_request.Request) -> None:
    """
    Generate a request_id for the Sanic request.

    :param request: Sanic request
    """
    request_id_var.set(
        gen_trace_id(request.path)
    )


class Sanic(vendor_sanic_app.Sanic):
    """
    Eventy modified Sanic application class.

    Includes eventy logging, specific sanic.access logging, and trace ids middlewares.
    """

    def __init__(
        self,
        name: str,
        **kwargs,
    ) -> None:
        """
        Initialize a Sanic application.

        Same parameters as sanic.Sanic, except ``log_config`` (set to None) and ``configure_logging`` (set to False)
        """
        if 'log_config' in kwargs:
            logger.warning(f"Sanic argument \"log_config\" will be overridden by eventy and set to None.")
        kwargs['log_config'] = None
        if 'configure_logging' in kwargs:
            logger.warning(f"Sanic argument \"configure_logging\" will be overridden by eventy and set to False.")
        kwargs['configure_logging'] = None
        super().__init__(
            name=name,
            **kwargs,
        )
        logging.getLogger('sanic.access').addFilter(sanic_access_log_msg_provider)
        logging.getLogger('sanic.access').addFilter(sanic_access_log_ignore_health_filter)
        self.register_middleware(sanic_correlation_id_provider, attach_to='request')
        self.register_middleware(sanic_request_id_provider, attach_to='request')



import logging
from io import StringIO

from django.http import HttpResponse
from django.test import TestCase

import eventy.config
import eventy.config.django
from eventy.logging import SimpleHandler


class TestCorrelationId(TestCase):

    def test_hello_generate_correlation_id(self):
        response: HttpResponse = self.client.get('/hello')
        self.assertEqual(response.status_code, 200)
        self.assertRegex(response.json()['hello'], "eventy-django-test:GET_/hello:.*")

    def test_hello_propagate_correlation_id(self):
        """
        NOTE: I didn't find any way to pass arbitrary headers to self.client (django.test.client.Client)

        See: run_integration_test.sh
        """

    def test_health(self):
        response = self.client.get('/health')
        self.assertEqual(response.status_code, 200)


class TestAccessLogging(TestCase):
    def setUp(self) -> None:
        self.log_string_io = StringIO()
        self.log_handler = SimpleHandler(colored=False, level='INFO')
        self.log_handler.stream = self.log_string_io
        logging.getLogger().addHandler(self.log_handler)

    def test_health_enable_logging(self):
        eventy.config.django.DJANGO_ACCESS_DISABLE_HEALTH_LOGGING = False

        self.client.get('/health')
        self.client.get('/hello')
        self.assertRegex(self.log_string_io.getvalue(), 'Request: GET /health')
        self.assertRegex(self.log_string_io.getvalue(), 'Response: GET /health')
        self.assertRegex(self.log_string_io.getvalue(), 'Request: GET /hello')
        self.assertRegex(self.log_string_io.getvalue(), 'Response: GET /hello')

    def test_health_disable_logging(self):
        eventy.config.django.DJANGO_ACCESS_DISABLE_HEALTH_LOGGING = True

        self.client.get('/health')
        self.client.get('/hello')
        self.assertNotRegex(self.log_string_io.getvalue(), 'Request: GET /health')
        self.assertNotRegex(self.log_string_io.getvalue(), 'Response: GET /health')
        self.assertRegex(self.log_string_io.getvalue(), 'Request: GET /hello')
        self.assertRegex(self.log_string_io.getvalue(), 'Response: GET /hello')




# Copyright (c) Qotto, 2021

"""
Eventy django integration configuration
"""

__all__ = [
    'DJANGO_ACCESS_DISABLE_HEALTH_LOGGING',
    'DJANGO_ACCESS_HEALTH_ROUTE',
]

DJANGO_ACCESS_DISABLE_HEALTH_LOGGING: bool = True
"""
Disable django access logs for health route
"""

DJANGO_ACCESS_HEALTH_ROUTE: str = '/health'
"""
Route for health checks
"""



from django.http import HttpResponse, JsonResponse
from django.views.decorators.http import require_GET

from eventy.trace_id import correlation_id_var


@require_GET
def health(request):
    return HttpResponse('OK')


def hello(request):
    return JsonResponse({'hello': correlation_id_var.get()})



# Copyright (c) Qotto, 2021

"""
Simple logging handler

>>> import sys
>>> import logging
>>> logger = logging.getLogger('my_logger')
>>> logger.setLevel('DEBUG')
>>> handler = SimpleHandler(stream=sys.stdout)
>>> logger.addHandler(handler)
>>> logger.info('test')
[...] [CID:] [RID:] INFO my_logger (<doctest eventy.logging L1) test
"""

from logging import Formatter, StreamHandler

from coloredlogs import ColoredFormatter

from .info_providers import correlation_id_provider, request_id_provider

__all__ = [
    'SimpleHandler',
]


class SimpleHandler(StreamHandler):
    """
    Simple logging handler, extending StreamHandler with optional colors and context information
    """

    def __init__(
        self,
        stream=None,
        fmt='%(asctime)s [CID:%(correlation_id)s] [RID:%(request_id)s] '
            '%(levelname)s %(name)s (%(module)s L%(lineno)d) '
            '%(message)s',
        datefmt='[%Y-%m-%d %H:%M:%S %z]',
        colored=False,
        level='DEBUG',
    ) -> None:
        """
        Initialize the handler

        :param stream: output stream, default None (sys.stderr), propagated to StreamHandler.__init__
        :param fmt: Format string used to initialize the :obj:`logging.Formatter`, can use the ``correlation_id`` and ``request_id`` fields of the log record
        :param datefmt: Date format string used to initialized the :obj:`logging.Formatter`
        :param colored: Output colored messages (different colors for log levels)
        :param level: log level, default 'DEBUG'. A log record is printed if its level is above both Handler level and Logger level
        """
        super().__init__(stream=stream)
        self.addFilter(correlation_id_provider)
        self.addFilter(request_id_provider)
        if colored:
            self.setFormatter(ColoredFormatter(fmt=fmt, datefmt=datefmt))
        else:
            self.setFormatter(Formatter(fmt=fmt, datefmt=datefmt))
        self.setLevel(level=level)



# Copyright (c) Qotto, 2021

"""
Record serialization errors
"""

__all__ = [
    'SerializationError',
    'UnknownRecordTypeError',
]


class SerializationError(Exception):
    """
    Base for all serialization errors
    """


class UnknownRecordTypeError(SerializationError):
    """
    Record type unknown (not EVENT, REQUEST, or RESPONSE)
    """

    def __init__(self, record_type: str) -> None:
        super().__init__(f'Unknown record type {record_type}')



# Copyright (c) Qotto, 2021

"""
Kafka implementation of the :mod:`eventy.messaging.transactional` API
"""

import logging

from confluent_kafka import Producer as KProducer, KafkaError

from .kafka_base import KafkaConsumer
from .kafka_errors import KafkaProducerCreationError, KafkaProducerTransactionError, KafkaConsumerCreationError
from eventy.messaging import TransactionalProducer, TransactionalProcessor
from eventy.record import Record
from eventy.serialization import ContextRecordSerializer, RecordSerializer

logger = logging.getLogger(__name__)

__all__ = [
    'KafkaTransactionalProcessor',
    'KafkaTransactionalProducer',
]


class KafkaTransactionalProducer(TransactionalProducer):

    def __init__(
        self,
        producer_config: dict[str, str],
        context_serializer: ContextRecordSerializer,
    ) -> None:
        self._context_serializer = context_serializer
        self._current_transaction: list[tuple[str, Record]] = list()
        try:
            # TODO: check sensible config (e.g. transactional.id) AND/OR provide sensible defaults
            self._transactional_producer = KProducer(producer_config)
            self._transactional_producer.init_transactions()
        except Exception as e:
            raise KafkaProducerCreationError from e

    def add_to_transaction(self, destination: str, record: Record, *args, **kwargs) -> None:
        self._current_transaction.append((destination, record))

    def commit(self):
        try:
            self._start_transaction()
            self._send_transaction_records()
            self._end_transaction()
        except Exception as e:
            raise KafkaProducerTransactionError from e

    def _start_transaction(self):
        self._transactional_producer.begin_transaction()

    def _send_transaction_records(self):
        for topic, record in self._current_transaction:
            message = self._context_serializer.encode(topic, record)
            self._transactional_producer.produce(topic, message, key=record.partition_key)

    def _end_transaction(self):
        if self._current_transaction:
            self._transactional_producer.commit_transaction()
            self._current_transaction.clear()


class KafkaTransactionalProcessor(KafkaTransactionalProducer, KafkaConsumer, TransactionalProcessor):

    def __init__(
        self,
        consumer_config: dict[str, str],
        consumer_topics: list[str],
        consumer_serializer: RecordSerializer,
        producer_config: dict[str, str],
        context_serializer: ContextRecordSerializer,
    ) -> None:
        try:
            KafkaConsumer.__init__(
                self,
                consumer_config,
                consumer_topics,
                consumer_serializer,
            )
            KafkaTransactionalProducer.__init__(
                self,
                producer_config,
                context_serializer,
            )
        except (KafkaConsumerCreationError, KafkaProducerCreationError) as e:
            raise KafkaProducerCreationError from e

    def _send_transaction_offsets(self):
        self._transactional_producer.send_offsets_to_transaction(
            self._kafka_consumer.position(self._kafka_consumer.assignment()),
            self._kafka_consumer.consumer_group_metadata(),
        )

    def commit(self):
        try:
            self._start_transaction()
            self._send_transaction_records()
            self._send_transaction_offsets()
        except Exception as e:
            # TODO: Some should be handled?
            # start can fail because it was already started on restart
            raise KafkaProducerTransactionError("Transaction could not start") from e
        try:
            self._end_transaction()
        except KafkaError as e:
            self._transactional_producer.abort_transaction()
            self._transactional_producer.begin_transaction()
            raise KafkaProducerTransactionError("Transaction aborted") from e



# Copyright (c) Qotto, 2021

"""
Event record
"""

from .record import Record

__all__ = [
    'Event',
]


class Event(Record):
    """
    Event implementation of the Record abstract base class
    """

    @property
    def type(self):
        """
        Record type (EVENT)

        :return: "EVENT"
        """
        return 'EVENT'



# Copyright (c) Qotto, 2021

"""
Response record
"""

from datetime import datetime
from typing import Dict, Any, Union, Optional
from uuid import UUID, uuid4

from urnparse import URN8141, InvalidURNFormatError

from .record import Record
from .record_errors import RecordAttributeTypeError, RecordAttributeValueError

__all__ = [
    'Response',
]


class Response(Record):

    def __init__(
        self,
        protocol_version: str,
        schema: str,
        version: str,
        source: str,
        destination: str,
        request_uuid: str,
        uuid: str = None,
        ok: bool = True,
        error_code: int = None,
        error_message: str = None,
        correlation_id: str = None,
        partition_key: str = None,
        date: datetime = None,
        context: Dict[str, Dict[str, Any]] = None,
        data: Dict[str, Any] = None
    ) -> None:
        """
        Initialize Record attributes

        :param protocol_version: SemVer version of the Eventy protocol
        :param schema: URN of the record schema
        :param version: SemVer version of the record schema
        :param source: URN of the record emitter
        :param destination: URN of the request destination
        :param request_uuid: UUID v4 of the corresponding request
        :param uuid: UUID v4 of the response. Default: auto-generated
        :param ok:
        :param error_code:
        :param error_message:
        :param correlation_id: Trace / correlation ID for trace_id. Default: auto-generated
        :param partition_key: Partition or database key. Default: None
        :param date: Date of the event. Default: now
        :param context: Keys should be URN of a service using the context
        :param data: Domain data

        :raises RecordAttributeError:
        """
        super().__init__(
            protocol_version,
            schema,
            version,
            source,
            uuid,
            correlation_id,
            partition_key,
            date,
            context,
            data,
        )
        self.destination = destination
        self.request_uuid = request_uuid
        self.ok = ok
        self.error_code = error_code
        self.error_message = error_message

    @property
    def type(self) -> str:
        return 'RESPONSE'

    @property
    def destination(self) -> str:
        return self._destination

    @destination.setter
    def destination(self, destination: str):
        try:
            URN8141.from_string(destination)
        except InvalidURNFormatError:
            raise RecordAttributeValueError('destination', destination, 'Not in URN format')
        self._destination = destination

    @property
    def request_uuid(self) -> str:
        return self._request_uuid

    @request_uuid.setter
    def request_uuid(self, request_uuid: str):
        if request_uuid is None:
            request_uuid = str(uuid4())
        try:
            UUID(request_uuid)
        except TypeError as e:
            raise RecordAttributeValueError('request_uuid', request_uuid, str(e)) from e
        except ValueError as e:
            raise RecordAttributeValueError('request_uuid', request_uuid, str(e)) from e
        self._request_uuid = request_uuid

    @property
    def ok(self) -> bool:
        return self._ok

    @ok.setter
    def ok(self, ok: Union[bool, str]):
        if isinstance(ok, bool):
            self._ok = ok
        elif isinstance(ok, str):
            if ok.lower() == 'true':
                self._ok = True
            elif ok.lower() == 'false':
                self._ok = False
            else:
                raise RecordAttributeValueError('ok', bool, ok)
        else:
            raise RecordAttributeTypeError('ok', bool, ok)

    @property
    def error_code(self) -> Optional[int]:
        return self._error_code

    @error_code.setter
    def error_code(self, error_code):
        # TODO: Check if ok is True? This would impose to update ok before error_code, but why not?
        if error_code is not None and not isinstance(error_code, int):
            raise RecordAttributeTypeError('error_code', 'int or None', error_code)
        self._error_code = error_code

    @property
    def error_message(self) -> Optional[str]:
        return self._error_message

    @error_message.setter
    def error_message(self, error_message):
        # TODO: Check if ok is True? This would impose to update ok before error_message, but why not?
        if error_message is not None and not isinstance(str, error_message):
            raise RecordAttributeTypeError('error_message', 'str or None', error_message)
        self._error_message = error_message



import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..', '..', '..'))
sys.path.append(os.path.join(os.path.dirname(__file__)))

from eventy.trace_id import correlation_id_var
from eventy.integration.sanic import Sanic
from sanic.response import text

app = Sanic("My Hello, world app")


@app.route('/hello')
async def hello(request):
    return text(correlation_id_var.get())


@app.route('/health')
async def health(request):
    return text('ok')


if __name__ == '__main__':
    app.run()



# Copyright (c) Qotto, 2021

"""
Include support for context bound trace ids

* :obj:`correlation_id_var` a Correlation ID to correlate multiple services
  involved in the same business operation
* :obj:`request_id_var` a Request ID to identify a single request on a service

Include support to generate trace ids

* :obj:`generator.gen_trace_id` trace id generator from a function
"""

from contextvars import ContextVar

__all__ = [
    'correlation_id_var',
    'request_id_var',
]

correlation_id_var: ContextVar[str] = ContextVar('correlation_id', default='')
"""
correlation_id context variable

>>> from eventy.trace_id import correlation_id_var
>>> correlation_id_var.set('my_operation_using_many_services')
>>> print(correlation_id_var.get())
'my_operation_using_many_services'
"""

request_id_var: ContextVar[str] = ContextVar('request_id', default='')
"""
request_id trace context variable

>>> from eventy.trace_id import request_id_var
>>> request_id_var.set('my_service_request')
>>> print(request_id_var.get())
'my_service_request'
"""



import logging
from enum import Enum

from environs import Env

env = Env()
env.read_env()

logger = logging.getLogger(__name__)

# Logging settings
LOG_LEVEL: str = env.enum(
    'LOG_LEVEL',
    type=Enum('LOG_LEVEL', 'DEBUG INFO WARNING ERROR CRITICAL'),
    ignore_case=True,
    default='INFO',
).name
GKE_LOGGING: bool = env.bool('GKE_LOGGING', False)
COLORED_LOGGING: bool = env.bool('COLORED_LOGGING', False)
SKIP_HEALTHCHECK_LOGGING: bool = env.bool('SKIP_HEALTHCHECK_LOGGING', False)
